@@ -2,11 +2,11 @@
 
 use std::fs;
 
-use cargo_test_support::prelude::*;
+use crate::prelude::*;
 use cargo_test_support::{
-    basic_manifest, git, main_file, project,
+    Execs, basic_manifest, git, main_file, project,
     registry::{Package, RegistryBuilder},
-    str, Execs,
+    str,
 };
 
 #[cargo_test]
